#include "MeshLoader.h"
#include <algorithm>
using namespace std;

std::ostream& operator << (std::ostream& out, const std::vector<Node>& elem) {
	for (auto i = elem.begin(); i < elem.end(); i++) {
		out << "   ID node: " << i->ID_Node << std::endl;
		out << "   Coordinates: ";
		for (auto j = i->Coordin.begin(); j < i->Coordin.end(); j++) {
			out << *j << " ";
		}
		out << std::endl << "   Flag: " << i->flag << std::endl << std::endl;
	}
	return out;
}

std::ostream& operator << (std::ostream& out, const std::vector <FiniteElement>& elem) {
	for (auto i = elem.begin(); i < elem.end(); i++) {
		out << "   ID_KE: " << i->ID_KE << std::endl;
		out << "   ID_Material: " << i->ID_Material << std::endl;
		out << "   ID_Nodes: ";
		for (auto j = i->ID_Nodes.begin(); j < i->ID_Nodes.end(); j++) {
			out << *j << " ";
		}
		out << std::endl << std::endl;
	}
	return out;
}

std::ostream& operator << (std::ostream& out, const std::vector <BoundaryFiniteElement>& elem) {
	for (auto i = elem.begin(); i < elem.end(); i++) {
		out << "   ID_Element: " << i->ID_Element << std::endl;
		out << "   ID_Border: " << i->ID_Border << std::endl;
		out << "   ID_Nodes: ";
		for (auto j = i->ID_Nodes.begin(); j < i->ID_Nodes.end(); j++) {
			out << *j << " ";
		}
		out << std::endl << std::endl;
	}
	return out;
}

std::ostream& operator << (std::ostream& out, const std::vector <vector<Node>>& elem) {
	int n = 1;
	for (auto i = elem.begin(); i < elem.end(); i++) {
		out << "Container_" << n << endl;
		out << *i;
		++n;
	}
	return out;
}

std::vector<Node> MeshLoader::vector_Node() { //!!! Лишнее копирование при возврате
	return MeshLoader::p_Node; //!!! Зачем  MeshLoader:: писать?
}

std::vector<FiniteElement> MeshLoader::vector_FiniteElement() { //!!! Лишнее копирование при возврате
	return MeshLoader::p_FiniteElement;  //!!! Зачем  MeshLoader:: писать?
}

std::vector<BoundaryFiniteElement> MeshLoader::vector_BoundaryFiniteElement() { //!!! Лишнее копирование при возврате
	return MeshLoader::p_BoundaryFiniteElement;  //!!! Зачем  MeshLoader:: писать?
}

//!!! Все контейнеры-результаты должны возвращать ID, а не сами КЭ и узлы.

std::vector<FiniteElement> MeshLoader::FiniteElementByThreeNodes(int id1, int id2, int id3) {
	std::vector<FiniteElement> result;
	auto a = p_FiniteElement.begin();

	while (a != p_FiniteElement.end()) {
		a = find_if(a, p_FiniteElement.end(), [id1, id2, id3](FiniteElement s) {
			return (find(s.ID_Nodes.begin(), s.ID_Nodes.end(), id1) != s.ID_Nodes.end() &&
				find(s.ID_Nodes.begin(), s.ID_Nodes.end(), id2) != s.ID_Nodes.end() &&
				find(s.ID_Nodes.begin(), s.ID_Nodes.end(), id3) != s.ID_Nodes.end());
			});
		if (a != p_FiniteElement.end()) {
			result.push_back(*a);
			a++;
		}
	}

	return result;

}

std::vector<FiniteElement> MeshLoader::FiniteElementByTwoNodes(int id1, int id2) {
	std::vector<FiniteElement> result;
	auto a = p_FiniteElement.begin();

	while (a != p_FiniteElement.end()) {
		a = find_if(a, p_FiniteElement.end(), [id1, id2](FiniteElement s) { //!!! Копирование s. Имена должны быть осмысленные
			return (find(s.ID_Nodes.begin(), s.ID_Nodes.end(), id1) != s.ID_Nodes.end() &&
			   	    find(s.ID_Nodes.begin(), s.ID_Nodes.end(), id2) != s.ID_Nodes.end());
			});
		if (a != p_FiniteElement.end()) {
			result.push_back(*a);
			a++;
		}
	}

	return result;

}

std::vector<Node> MeshLoader::BoundaryNodes(int id) {
	std::vector<Node> result;
	
	//!!! Реализация везде должна быть на основе алгоритмов
	
	for (auto a = p_BoundaryFiniteElement.begin(); a < p_BoundaryFiniteElement.end(); a++) {
		if (a->ID_Border == id) {
			for (auto b = a->ID_Nodes.begin(); b < a->ID_Nodes.end(); b++) {
			    
			    //!!! Не эффективная реализация и неверный тип контейнера для предотвращения дубликатов.
				if (find_if(result.begin(), result.end(), [b](Node s) { //!!! Копирование s
					return (s.ID_Node == *b);
					}) == result.end()) {
					result.push_back(p_Node.at((*b)-1));
				}
			}
		}
	}

	return result;
}

std::vector<FiniteElement> MeshLoader::KE(int id) { //!!! Из названия не понятно, что метод делает и что за id 
	std::vector<FiniteElement> result;

    //!!! У полей класса префикс m_, а не p_
	for (auto a = p_BoundaryFiniteElement.begin(); a < p_BoundaryFiniteElement.end(); a++) {
		if (a->ID_Border == id) {
		    //!!! Для каждого граничено КЭ обходятся все тетраэдры? Что тут делается?
		    
			for (auto b = p_FiniteElement.begin(); b < p_FiniteElement.end(); b++) {
				if (find_if(a->ID_Nodes.begin(), a->ID_Nodes.end(), [b](auto s) { //!!! Копирование s
				
					return (find(b->ID_Nodes.begin(), b->ID_Nodes.end(), s) == b->ID_Nodes.end());
					}) == a->ID_Nodes.end()
				    && find_if(result.begin(), result.end(), [b](auto s) {
					return (s.ID_KE==b->ID_KE);
					}) == result.end()) {
					result.push_back(*b);
				}
			}
		}
	}
	return result;
}

std::vector<BoundaryFiniteElement> MeshLoader::BoundaryKE(int id) {
	std::vector<BoundaryFiniteElement> result;

    //!!! Где алгоритмы?

	for (auto a = p_BoundaryFiniteElement.begin(); a < p_BoundaryFiniteElement.end(); a++) {
		if (a->ID_Border == id) {
			result.push_back(*a);
		}
	}

	return result;
}

void MeshLoader::NewNodes() {
	auto n = p_Node.size();
	for (int i = 0; i < n - 1; i++) {
		for (int j = 1 + i; j < n; j++) {
			auto a = &p_Node.at(i);
			auto b = &p_Node.at(j);
			
			//!!! Крайне не эффективно. Не выполнено услови задания по сложности.
			//!!! Просмотренные ребра должны хешироваться. Должны быть проверка, было ли просмотрено ребро и был ли создан узел?
			//!!! Если да, то изъять ID вставленного узла из ребра (третий параметр). Если нет - то создать узел.
			if (find_if(p_FiniteElement.begin(), p_FiniteElement.end(), [a, b](auto s) { //!!! Копирование s
				return (find(s.ID_Nodes.begin(), s.ID_Nodes.end(), a->ID_Node) != s.ID_Nodes.end() &&
					find(s.ID_Nodes.begin(), s.ID_Nodes.end(), b->ID_Node) != s.ID_Nodes.end());
				}) != p_FiniteElement.end()) {
				//  
				Node c;
				c.flag = false;
				c.ID_Node = p_Node.size() + 1;
				for (int i = 0; i < a->Coordin.size(); i++) {
					c.Coordin.push_back((a->Coordin.at(i) + b->Coordin.at(i)) / 2);
				}
				p_Node.push_back(c);
				auto c1 = p_FiniteElement.begin();
				while (c1 != p_FiniteElement.end()) {
					c1 = find_if(c1, p_FiniteElement.end(), [a, b](auto s) { //!!! Копирование s
						return (find(s.ID_Nodes.begin(), s.ID_Nodes.end(), a->ID_Node) != s.ID_Nodes.end() &&
							find(s.ID_Nodes.begin(), s.ID_Nodes.end(), b->ID_Node) != s.ID_Nodes.end());
						});
					if (c1 != p_FiniteElement.end()) {
						c1->ID_Nodes.push_back(c.ID_Node);
						c1++;
					}
				}
				auto c2 = p_BoundaryFiniteElement.begin();
				while (c2 != p_BoundaryFiniteElement.end()) {
					c2 = find_if(c2, p_BoundaryFiniteElement.end(), [a, b](auto s) {
						return (find(s.ID_Nodes.begin(), s.ID_Nodes.end(), a->ID_Node) != s.ID_Nodes.end() &&
							find(s.ID_Nodes.begin(), s.ID_Nodes.end(), b->ID_Node) != s.ID_Nodes.end());
						});
					if (c2 != p_BoundaryFiniteElement.end()) {
						c2->ID_Nodes.push_back(c.ID_Node);
						c2++;
					}
				}
			}
		}
	}
}

std::vector<std::vector<Node>> MeshLoader::ContainerOfNodeContainers() { //!!! Непонятное название
	std::vector<std::vector<Node>> result; //!!! Неверный тип контейнера
	for (auto a = p_Node.begin(); a < p_Node.end(); a++) {
		std::vector<Node> NeighboringNodesOfNode_a;
		for (auto b = p_FiniteElement.begin(); b < p_FiniteElement.end(); b++) {
			if (find(b->ID_Nodes.begin(), b->ID_Nodes.end(), a->ID_Node)!=b->ID_Nodes.end()) {
				for (auto c = b->ID_Nodes.begin(); c < b->ID_Nodes.end(); c++) {
				    
				    //!!! Не эффективно
					if (find_if(NeighboringNodesOfNode_a.begin(), NeighboringNodesOfNode_a.end(), [c](Node s) { //!!! Копирование s
						return (s.ID_Node == *c);
						}) == NeighboringNodesOfNode_a.end() && *c!=a->ID_Node) {
						NeighboringNodesOfNode_a.push_back(p_Node.at(*c-1));
					}
				}
			}
		}
		result.push_back(NeighboringNodesOfNode_a);
	}
	return result;
}

void MeshLoader::print_Node() {
	cout << p_Node;
}

void MeshLoader::print_FiniteElement() {
	cout << p_FiniteElement;
}

void MeshLoader::print_BoundaryFiniteElement() {
	cout << p_BoundaryFiniteElement;
}